#include "../SDK/foobar2000.h"
#include "../helpers/helpers.h"
#include "resource.h"

/**********************************
This file contains the implementation of two DSP classes.
The first one is an artificial example that demonstrates
chunk handling and some other general features of DSPs.
The second one is adapted from Marc Heubeck's
"Noise sharpening" DSP and focuses on handling per-instance
DSP configuration.
**********************************/

DECLARE_COMPONENT_VERSION(
	"DSP Tutorial",
	"1.0.2",
	"DSP Tutorial for foobar2000 0.9.5\n"
	"Written by Holger Stenger\n"
	"'Tutorial DSP with presets' derived from foo_dsp_delta with kind permission from Marc Heubeck."
	);

/**********************************
DSP without presets
**********************************/

// This DSP does not alter the signal.
// It merely mangles audio chunks to demonstrate some features
// of the DSP service class. It will buffer one chunk at a time,
// thereby introducing latency in the DSP chain (and some overhead
// for copying data around).
class dsp_tutorial_nopreset : public dsp_impl_base
{
	bool m_chunk_valid;
	audio_chunk_impl m_chunk;

public:
	dsp_tutorial_nopreset() {
		// Mark buffer as empty.
		m_chunk_valid = false;
	}

	// Every DSP type is identified by a GUID.
	static GUID g_get_guid() {
		// Create these with guidgen.exe.
		static const GUID guid = { 0x22ba0124, 0x5ddb, 0x42ee, { 0xb5, 0x2a, 0xc9, 0x33, 0x89, 0x5a, 0xf0, 0x99 } };
		
		return guid;
	}

	// We also need a name, so the user can identify the DSP.
	// The name we use here does not describe what the DSP does,
	// so it would be a bad name. We can excuse this, because it
	// doesn't do anything useful anyway.
	static void g_get_name(pfc::string_base & p_out) {
		p_out = "Tutorial DSP without presets";
	}

	virtual void on_endoftrack(abort_callback & p_abort) {
		// This method is called when a track ends.
		// We need to do the same thing as flush(), so we just call it.
		flush();
	}

	virtual void on_endofplayback(abort_callback & p_abort) {
		// This method is called on end of playback instead of flush().
		// We need to do the same thing as flush(), so we just call it.
		flush();
	}

	// The framework feeds input to our DSP using this method.
	// Each chunk contains a number of samples with the same
	// stream characteristics, i.e. same sample rate, channel count
	// and channel configuration.
	virtual bool on_chunk(audio_chunk * chunk, abort_callback & p_abort) {
		// Output the buffered chunk.
		if (m_chunk_valid) {
			audio_chunk * out = insert_chunk(m_chunk.get_sample_count());
			out->copy(m_chunk);
		}

		// Copy the input chunk to our buffer.
		m_chunk.copy(*chunk);
		m_chunk_valid = true;

		// Return false to keep the input chunk from being added to the output.
		return false;
	}

	virtual void flush() {
		// Clear the buffer.
		m_chunk.reset();
		m_chunk_valid = false;
	}

	virtual double get_latency() {
		// If the buffered chunk is valid, return its length.
		// Otherwise return 0.
		return m_chunk_valid ? m_chunk.get_duration() : 0.0;
	}

	virtual bool need_track_change_mark() {
		// Return true if you need to know exactly when a new track starts.
		// Beware that this may break gapless playback, as at least all the
		// DSPs before yours have to be flushed.
		// To picture this, consider the case of a reverb DSP which outputs
		// the sum of the input signal and a delayed copy of the input signal.
		// In the case of a single track:

		// Input signal:   01234567
		// Delayed signal:   01234567

		// For two consecutive tracks with the same stream characteristics:

		// Input signal:   01234567abcdefgh
		// Delayed signal:   01234567abcdefgh

		// If the DSP chain contains a DSP that requires a track change mark,
		// the chain will be flushed between the two tracks:

		// Input signal:   01234567  abcdefgh
		// Delayed signal:   01234567  abcdefgh
		return false;
	}
};

// We need a service factory to make the DSP known to the system.
// DSPs use special service factories that implement a static dsp_entry
// that provides information about the DSP. The static methods in our
// DSP class our used to provide the implementation of this entry class.
// The entry is used to instantiate an instance of our DSP when it is needed.
// We use the "nopreset" version of the DSP factory which blanks out
// preset support.
// Note that there can be multiple instances of a DSP which are used in
// different threads.
static dsp_factory_nopreset_t<dsp_tutorial_nopreset> foo_dsp_tutorial_nopreset;


/**********************************
DSP with presets
**********************************/


// We will use a helper struct to held the configuration for our
// DSP which can be used to pass around the relevant data to
// the configuration dialog. The struct also implements the 
// serialisation from and to dsp_preset instances which store
// configuration data across sessions.
struct t_dsp_tutorial_params
{
	// The strength of the effect in percent.
	t_int32 m_percent;

	// The constructor sets the default values.
	t_dsp_tutorial_params(t_int32 p_percent = 100) : m_percent(p_percent) {
	}

	// The GUID that identifies this DSP and its configuration.
	static const GUID &g_get_guid() {
		static const GUID guid = { 0x7f6baf16, 0x1b65, 0x40a4, { 0x09, 0x99, 0x73, 0x60, 0xa7, 0x53, 0x73, 0x6f } };

		return guid;
	}

	// Read data from a preset.
	bool set_data(const dsp_preset & p_data) {
		if (p_data.get_owner() != g_get_guid()) return false;
		if (p_data.get_data_size() != sizeof(t_int32)) return false;
		t_int32 temp = *(t_int32 *)p_data.get_data();
		byte_order::order_le_to_native_t(temp);
		m_percent = temp;
		return true;
	}

	// Write data to a preset.
	void get_data(dsp_preset & p_data) {
		p_data.set_owner(g_get_guid());
		t_int32 temp = m_percent;
		byte_order::order_native_to_le_t(temp);
		p_data.set_data(&temp, sizeof(temp));
	}
};

// Our configuration dialog is implemented as a modal dialog.
// The helper class implements some safety measures to avoid
// opening multiple non-nested modal dialogs.
class dialog_dsp_tutorial : public dialog_helper::dialog_modal
{
public:
	dialog_dsp_tutorial(const dsp_preset & p_data, dsp_preset_edit_callback & p_callback) : m_old_data(p_data), m_callback(p_callback), m_dirty(false) {
		m_params.set_data(m_old_data);
	}

	virtual BOOL on_message(UINT msg, WPARAM wp, LPARAM lp) {
		switch(msg) {
		case WM_INITDIALOG:
			{
				HWND slider = GetDlgItem(get_wnd(), IDC_SLIDER);
				// Needs redraw flag set for both calls, or things will be messed
				// up if m_percent is zero.
				SendMessage(slider, TBM_SETRANGE, TRUE, MAKELONG(-50, 200));
				SendMessage(slider, TBM_SETPOS, TRUE, m_params.m_percent);
				SendMessage(slider, TBM_SETTICFREQ, 25, 0);
				update_display();
			}
			break;

			// Slider has been moved.
		case WM_HSCROLL:
			{
				m_dirty = true;
				update_display();
			}
			break;

		case WM_COMMAND:
			switch (wp) {
			case IDOK:
				{
					end_dialog(1);
				}
				break;

			case IDCANCEL:
				{
					m_callback.on_preset_changed(m_old_data);
					end_dialog(0);
				}
				break;
			}
			break;
		}
		return 0;
	}

private:
	void update_display() {
		m_params.m_percent = SendDlgItemMessage(get_wnd(), IDC_SLIDER, TBM_GETPOS, 0, 0);
		if (m_dirty) {
			dsp_preset_impl data;
			m_params.get_data(data);
			m_callback.on_preset_changed(data);
			m_dirty = false;
		}

		uSetDlgItemText(get_wnd(), IDC_STATIC_DISPLAY,
			pfc::string_formatter() << pfc::format_int(m_params.m_percent) << "%");
	}

	bool m_dirty;
	const dsp_preset & m_old_data;
	t_dsp_tutorial_params m_params;
	dsp_preset_edit_callback & m_callback;
};

// Finally, this is our DSP class. We need a few extra methods
// to support presets.
class dsp_tutorial : public dsp_impl_base
{
	pfc::array_t<audio_sample> m_last_sample;
	audio_sample m_factor;

public:
	dsp_tutorial(const dsp_preset & p_data) : m_factor(0.0) {
		set_data(p_data);
	}

	static GUID g_get_guid() {
		return t_dsp_tutorial_params::g_get_guid();
	}

	static void g_get_name(pfc::string_base & p_out) {
		p_out = "Tutorial DSP with presets";
	}

	// Return if we have a configuration popup (we do).
	static bool g_have_config_popup() {
		return true;
	}

	// Show our configuration popup.
	// The provided callback is used to report configuration changes back to the caller.
	static void g_show_config_popup(const dsp_preset & p_data, HWND p_parent, dsp_preset_edit_callback & p_callback) {
		t_dsp_tutorial_params params;
		if (params.set_data(p_data)) {
			dialog_dsp_tutorial dlg(p_data, p_callback);
			// TODO dialog_helper::dialog_modal is deprecated
			dlg.run(IDD_CONFIG, p_parent);
		}
	}

	// Return our default preset.
	static bool g_get_default_preset(dsp_preset & p_out) {
		t_dsp_tutorial_params().get_data(p_out);
		return true;
	}

	// Read parameters from the provided preset.
	bool set_data(const dsp_preset & p_data) {
		t_dsp_tutorial_params params;
		if (!params.set_data(p_data)) return false;
		m_factor = (audio_sample)(params.m_percent * 0.01);
		return true;
	}

	virtual void on_endoftrack(abort_callback & p_abort) {
		flush();
	}

	virtual void on_endofplayback(abort_callback & p_abort) {
		flush();
	}

	// Process chunk.
	virtual bool on_chunk(audio_chunk * chunk, abort_callback & p_abort) {
		unsigned channel_count = chunk->get_channels();
		t_size sample_count = chunk->get_sample_count();

		if (channel_count != m_last_sample.get_size()) {
			// number of channels has changed - (re)initialize "history"
			m_last_sample.set_size(channel_count);
			m_last_sample.fill<audio_sample>(0);
		}

		audio_sample * current = chunk->get_data();

		for (t_size sample = 0; sample < sample_count; sample++) {
			for (unsigned channel = 0; channel < channel_count; channel++) {
				audio_sample temp = current[0];
				audio_sample delta = temp - m_last_sample[channel];
				current[0] += delta * m_factor;
				m_last_sample[channel] = temp;
				current++;
			}
		}

		// Add (modified) input chunk to output.
		return true;
	}

	virtual void flush() {
		// Nothing to flush.
	}

	virtual double get_latency() {
		// We have no buffer, so latency is 0.
		return 0.0;
	}

	virtual bool need_track_change_mark() {
		return false;
	}
};

static dsp_factory_t<dsp_tutorial> foo_dsp_tutorial;
